#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <errno.h>

#define error_in_step()\
{\
	fprintf(stderr,"The function step became less than the minimum step \n"\
	"line %d function %s \n"\
	"Entered the new minimum step in step_rkf45 \n",\
	__LINE__,__FUNCTION__);\
	abort();\
}

typedef void (*model_callback_t)(float t0, float *y0, float *dy);

void model(float t0, float *y0, float *dy)
{
	assert(y0 != NULL && dy != NULL);

	dy[0] = y0[1];
	dy[1] = 2.0*y0[1] - y0[0];
}


static int
rkf45(float h, float t0, float *y0, float *dy, float *y, unsigned int n, model_callback_t model,float epsilon, float *q)
{
	
	float delta_y;
	int i;
	float *k;
	float *k1;
	float *k2;
	float *k3;
	float *k4;
	float *k5;
	float max_delta_y;

	*q = 0;
	max_delta_y = 0;

	k1 = alloca(sizeof(float)*n);
	k2 = alloca(sizeof(float)*n);
	k3 = alloca(sizeof(float)*n);
	k4 = alloca(sizeof(float)*n);
	k5 = alloca(sizeof(float)*n);
	k = alloca(sizeof(float)*n);

	
	for (i = 0; i < n; i++) {
		k1[i] = h*dy[i];
		k[i] = y0[i] + 0.25*k1[i];
	}
	
	(*model)(t0+0.25*h, k, dy);
	
	for (i = 0; i < n; i++){
		k2[i] = h*dy[i];
		k[i] = y0[i] + 3.0*k1[i]/32.0 + 9.0*k2[i]/32.0;
	}
	
	(*model)(t0 + 3.0*h / 8.0, k, dy);
	
	for (i = 0; i < n; i++) {
		k3[i] = h*dy[i];
		k[i] = y0[i] + (1932.0*k1[i] - 7200.0*k2[i] + 7296.0*k3[i])/2197.0;
	}

	(*model)(t0+12.0/13.0*h, k, dy);

	for (i = 0; i < n; i++) {
		k4[i] = h*dy[i];
		k[i] = y0[i] + 439.0/216.0*k1[i] - 8.0*k2[i] + 3680.0/513.0*k3[i] - 845.0/4104.0*k4[i];
	}

	(*model)(t0+h, k, dy);
	
	for (i = 0; i < n; i++) {
		k5[i] = h*dy[i];
		k[i] = y0[i] - 8.0/27.0*k1[i] + 2*k2[i] - 3544.0/2565.0*k3[i] + 1859.0/4104.0*k4[i] - 11.0/40.0*k5[i];
	}
	
	(*model)(t0+0.5*h, k, dy);

	for (i = 0; i < n; i++) {
		delta_y = 2.0/55.0*h*dy[i] + 1.0/360.0*k1[i] - 128.0/4275.0*k3[i] + (28.561/56.43 - 2197.0/4101.0)*k4[i] + 0.02*k5[i];
		max_delta_y = (max_delta_y < fabs(delta_y)) ? fabs(delta_y) : max_delta_y;
	}
	
	if (max_delta_y != 0)
		*q = 0.8409864*sqrt(sqrt(h*epsilon / max_delta_y));
	else
		*q = 5;

	for (i = 0; i < n; i++){
		y[i] = y0[i] + 25.0*k1[i]/216.0 + 1408.0/2565.0*k3[i] + 2197.0/4101.0*k4[i] - 0.2*k5[i];
	}
	for (i = 0; i < n; i++)
		y0[i] = y[i];
	return 0;
}

int 
step_rkf45(float *h, float t0, float *y0, float *dy, float *y, unsigned int n, float max_h, float min_h, float max_t, model_callback_t model,float epsilon)
{
	float q;
	while (t0 + *h <= max_t) { /*пока не достигнем конечной точки вызываем метод*/
	
		if (*h < min_h)
			error_in_step();
			
		model(t0, y0, dy);
		rkf45(*h, t0, y0, dy, y, n, model,epsilon, &q);
		
		if (q < 0.1 && *h * q > min_h)
			while (q < 0.1 && *h * q > min_h) {/*вызываем метод, пока q не станет нормальным или шаг минимальным*/
				*h = *h * q;
				if (*h < min_h)
					error_in_step();
				rkf45(*h, t0, y0, dy, y, n, model,epsilon, &q);
			}
			
		if (q > 4.0 && *h != max_h)
			while (q > 4.0 && *h != max_h) {/*вызываем метод, пока q не станет нормальным или шаг максимальным*/
				*h =  (*h * q < max_h) ? *h * q : max_h;
				rkf45(*h, t0, y0, dy, y, n, model,epsilon, &q);
			}
		
		
		t0 = t0 + *h;
	}
	
	if (t0 + *h > max_t ) {/*если мы перешагнули конечную точку*/
		model(t0, y0, dy);
		*h = t0 - max_t;
		t0 = max_t - *h;
		rkf45(*h, t0, y0, dy, y, n, model,epsilon, &q);
	}

	return 0;
	
}


int 
step_rkf45(float *h, float t0, float *y0, float *dy, float *y, unsigned int n, float max_h, float min_h, float max_t, model_callback_t model,float epsilon);
	
int main()
{
	float h;
	float t;
	float y0[2];
	float dy[2];
	float y[2];
	unsigned int n;
	
	t = 0.0;
	h = 0.0005;
	n = 2;

	/*initial values*/ 
	y0[0] = 0;
	y0[1] = 1;
	y[0] = 0;
	y[1] = 0;

	

	printf("y0[0] = %e y0[1] = %e\n", y0[0], y0[1]);

		step_rkf45(&h, t, y0, dy, y, n,  0.001, 0.0001, 1.0, model,0.0001);
		
			printf("t = %3.4e | y[0] = %4.3e | y[1] = %4.3e\n", t, y[0], y[1]);
	return 0;
}

